class HctefError(Exception):
    """Base exception for all hctef errors."""


class HctefNetworkError(HctefError, IOError):
    """Network-related error while fetching data."""


class HctefUrlError(HctefError, ValueError):
    """Invalid URL for file."""



from .http_file import HttpFile

try:
    from .__version__ import __version__, __version_tuple__
except ImportError:
    __version__ = '0.0.0'
    __version_tuple__ = ('0', '0', '0')

__all__: list[str] = [
    'HttpFile',
    '__version__',
    '__version_tuple__',
]



from __future__ import annotations

import urllib.request

from typing import Literal, Self

from .exceptions import HctefNetworkError, HctefUrlError
from .file_read_cache import FileReadCache


def _check_url(url: str) -> None:
    """
    Validate that URL is a valid HTTP/HTTPS URL.

    Args:
        url: URL to validate

    Raises:
        HctefUrlError: If URL doesn't start with http: or https:
    """
    if not url.startswith(('http:', 'https:')):
        raise HctefUrlError("URL must start with 'http:' or 'https:'")


class _OpenedHttpFile:
    """
    Internal class of HttpFile for managing state while open
    """

    def __init__(
        self,
        http_file: HttpFile,
    ) -> None:
        """
        Open the HTTP URL for the HttpFile.

        Args:
            http_file: An HttpFile instance.

        Raises:
            HctefNetworkError: If server doesn't support range requests
        """

        self.http_file = http_file
        self._position = 0
        self._size = self._get_file_size()
        self._minimum_range_request_bytes = min(
            self.http_file._minimum_range_request_bytes,
            self._size,
        )
        self._cache = FileReadCache(
            self._size,
            self._fetch_range,
            minimum_request_size=min(
                self.http_file._minimum_range_request_bytes,
                self._size,
            ),
        )

        prefetch_bytes = min(
            self.http_file._prefetch_bytes,
            self._size,
        )
        prefetch_direction = self.http_file._prefetch_direction
        if prefetch_bytes > 0 and prefetch_direction == 'START':
            self._cache.read(0, prefetch_bytes)
        elif prefetch_bytes > 0 and prefetch_direction == 'END':
            self._cache.read(self._size - prefetch_bytes, self._size)

    def _get_file_size(self) -> int:
        """
        Get total file size using HTTP range request.

        Returns:
            File size in bytes

        Raises:
            HctefNetworkError: If size cannot be determined
        """
        try:
            request = urllib.request.Request(  # noqa: S310
                self.http_file.url,
                headers={'Range': 'bytes=0-'},
            )
            with urllib.request.urlopen(request) as response:  # noqa: S310
                content_range = response.headers.get('Content-Range')
                if content_range:
                    return int(content_range.split('/')[-1])

                # If no Content-Range header, server doesn't support ranges
                raise HctefNetworkError(
                    f'Server does not support range requests for {self.http_file.url}',
                )
        except Exception as e:
            raise HctefNetworkError(
                f'Cannot determine file size for {self.http_file.url}',
            ) from e

    def read(self, size: int | None = None, /) -> bytes:
        if size is None:
            size = self._size - self._position

        if size < 0:
            raise ValueError(f'Cannot read negative number of bytes, got: {size}')

        if size == 0:
            return b''

        start = self._position
        end = min(start + size, self._size)

        data = self._cache.read(start, end)

        self._position = end
        return data

    def _fetch_range(self, start: int, end: int) -> bytes:
        """
        Fetch byte range using HTTP request and add to cache.

        Args:
            start: Start byte position (inclusive)
            end: End byte position (exclusive)

        Raises:
            HctefNetworkError: If range request fails
        """
        if start >= end or start < 0 or end > self._size:
            raise HctefUrlError(
                f'Invalid byte range: {start}-{end} (file size: {self._size})',
            )

        try:
            request = urllib.request.Request(  # noqa: S310
                self.http_file.url,
                headers={'Range': f'bytes={start}-{end - 1}'},
            )
            with urllib.request.urlopen(request) as response:  # noqa: S310
                return response.read()
        except Exception as e:
            raise HctefNetworkError(
                f'Failed to fetch bytes {start}-{end} from {self.http_file.url}:',
            ) from e

    def seek(self, offset: int, whence: int = 0, /) -> int:
        if whence == 0:  # Absolute position
            new_pos = offset
        elif whence == 1:  # Relative to current position
            new_pos = self._position + offset
        elif whence == 2:  # Relative to end
            new_pos = self._size + offset
        else:
            raise ValueError(f'Invalid whence value: {whence}')

        if new_pos < 0:
            new_pos = 0
        elif new_pos > self._size:
            new_pos = self._size

        self._position = new_pos
        return self._position

    def tell(self) -> int:
        return self._position


class HttpFile:
    """
    File-like wrapper for HTTP URLs with range request support.
    """

    def __init__(
        self,
        url: str,
        minimum_range_request_bytes: int = 8192,
        prefetch_bytes: int = 2**20,
        prefetch_direction: Literal['START', 'END'] = 'END',
    ) -> None:
        """
        Initialize HTTP file wrapper.

        Args:
            url: HTTP/HTTPS URL for a file

        Keyword Args:
            minimum_range_request_bytes:
                Least number of bytes to request,
                except when filling cache gaps
            prefetch_bytes:
                How many bytes to request when initializing the class.
                Set to 0 or less to disable prefetch. Default 1 MiB.
            prefetch_direction:
                Whether to prefetch from file start or file end.
                Possible values `START` or `END`.

        Raises:
            HctefUrlError: If URL is invalid
            HctefNetworkError: If server doesn't support range requests
        """

        _check_url(url)
        self.url = url
        self._prefetch_bytes = prefetch_bytes
        self._prefetch_direction = prefetch_direction
        self._minimum_range_request_bytes = minimum_range_request_bytes
        self._opened: _OpenedHttpFile | None = None

    @property
    def _ohf(self) -> _OpenedHttpFile:
        if not self._opened:
            raise ValueError('I/O operation on closed file')
        return self._opened

    def open(self) -> Self:
        self._opened = _OpenedHttpFile(self)
        return self

    def close(self) -> None:
        """
        Close the file (clears cache).
        """
        self._opened = None

    def __enter__(self) -> Self:
        return self.open()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def __repr__(self) -> str:
        if self._opened:
            return (
                f'HttpFile(url={self.url!r}, opened=True, '
                f'size={self._ohf._size}, pos={self._ohf._position})'
            )
        return f'HttpFile(url={self.url!r}, opened=False)'

    def read(self, size: int | None = None, /) -> bytes:
        """
        Read bytes from current position.

        Args:
            size: Number of bytes to read (-1 for all remaining)

        Returns:
            Bytes read from the file
        """
        return self._ohf.read(size)

    def seek(self, offset: int, whence: int = 0, /) -> int:
        """
        Change stream position.

        Args:
            offset: Byte offset
            whence: How to interpret offset (0=absolute, 1=relative, 2=from end)

        Returns:
            New absolute position
        """
        return self._ohf.seek(offset, whence)

    def tell(self) -> int:
        """
        Get current stream position.

        Returns:
            Current byte position in file
        """
        return self._ohf._position

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def seekable(self) -> bool:
        return True



import pytest

PARQUET_BASE_URL = (
    'https://raw.githubusercontent.com/apache/parquet-testing/master/data'
)


@pytest.fixture
def parquet_url(parquet_file_name: str) -> str:
    return f'{PARQUET_BASE_URL}/{parquet_file_name}.parquet'



import io
import logging

from collections.abc import Callable

from .interval_tree import BinningIntervalTree, Interval

logger = logging.getLogger(__name__)


class FileReadCache:
    """
    A memory-efficient class to manage caching chunks of a file.
    Typically used to cache data in a remote file to prevent
    many small reads, or repeated reads of the same bytes.
    """

    def __init__(
        self,
        file_size: int,
        fetch_range: Callable[[int, int], bytes],
        minimum_request_size: int = 8192,
    ) -> None:
        if file_size < 0:
            raise ValueError('File size cannot be less than zero')

        self.file_size = file_size
        self.cache = BinningIntervalTree[bytes](bin_size=minimum_request_size)
        self._fetcher = fetch_range
        self._minimum_request_size = minimum_request_size

    def clear(self):
        """
        Resets the cache, clearing all stored data and intervals.
        """
        logger.debug('Clearing all cached data')
        self.cache.clear()

    def _store_and_merge(self, start, end, chunk):
        """
        Stores a new chunk of data and merges it with any adjacent or
        overlapping chunks in the cache.
        """
        overlapping = self.cache.find_overlapping(start - 1, end + 1)
        min_start = start
        max_end = end
        parts = {start: chunk}

        for iv in overlapping:
            min_start = min(min_start, iv.begin)
            max_end = max(max_end, iv.end)
            parts[iv.begin] = iv.data
            self.cache.remove(iv)

        merged_data = io.BytesIO()
        for offset in sorted(parts.keys()):
            merged_data.write(parts[offset])

        self.cache.add(Interval(min_start, max_end, merged_data.getvalue()))
        logger.debug(
            '--- CACHE UPDATE: Cache now contains: %s',
            [(iv.begin, iv.end) for iv in sorted(self.cache)],
        )

    def _fetch_range(self, start: int, end: int) -> None:
        """
        Fetches a range of bytes, applying minimum request size logic
        in a cache-aware manner to avoid re-downloading.
        """
        more = self._minimum_request_size - (end - start)

        if more > 0:
            overlapping = self.cache.find_overlapping(
                end,
                min(self.file_size, end + self._minimum_request_size),
            )
            right_wall = overlapping[0].begin if overlapping else self.file_size
            end = min(end + more, right_wall)
            more = self._minimum_request_size - (end - start)

        if more > 0:
            overlapping = self.cache.find_overlapping(
                max(0, start - self._minimum_request_size),
                start,
                sort_by_end=True,
            )
            left_wall = overlapping[0].end if overlapping else 0
            start = max(start - more, left_wall)

        chunk = self._fetcher(start, end)
        self._store_and_merge(start, end, chunk)

    def _find_missing_ranges(self, start: int, end: int) -> list[Interval]:
        """
        A simple, robust method to find '''holes''' in the cache.
        """
        missing: list[Interval] = []
        relevant_intervals = sorted(self.cache.find_overlapping(start, end))
        pos = start

        for interval in relevant_intervals:
            if pos < interval.begin:
                missing.append(Interval(pos, interval.begin, None))
            pos = max(pos, interval.end)

        if pos < end:
            missing.append(Interval(pos, end, None))

        return missing

    def read(self, start: int, end: int) -> bytes:
        """
        Reads a range of bytes, utilizing the cache and fetching if necessary.
        """
        if end > self.file_size:
            raise ValueError('Read request extends beyond the end of the file.')

        logger.debug(
            'Read Request for bytes %s-%s',
            start,
            end - 1,
        )

        missing_intervals = self._find_missing_ranges(start, end)

        if not missing_intervals:
            logger.debug('CACHE HIT: All requested data already in cache.')
        else:
            logger.debug(
                'CACHE MISS: Missing intervals are: %s',
                [(iv.begin, iv.end) for iv in missing_intervals],
            )
            for interval in missing_intervals:
                # Check if the interval (or part of it) has been filled
                # by a previous larger fetch in this same read() call.
                still_missing = self._find_missing_ranges(
                    interval.begin,
                    interval.end,
                )
                for gap in still_missing:
                    self._fetch_range(gap.begin, gap.end)

        result_buffer = io.BytesIO()
        cached_chunks = sorted(self.cache.find_overlapping(start, end))

        for interval in cached_chunks:
            read_start = max(start, interval.begin)
            read_end = min(end, interval.end)

            slice_start = read_start - interval.begin
            slice_end = read_end - interval.begin

            result_buffer.write(interval.data[slice_start:slice_end])

        return result_buffer.getvalue()



import os

from collections.abc import Callable

import pytest

from hctef.file_read_cache import FileReadCache

type FetchLog = list[tuple[int, int]]
type FetchRange = Callable[[int, int], bytes]


@pytest.fixture
def fetch_log() -> FetchLog:
    return []


@pytest.fixture
def fetch_range(
    fetch_log: FetchLog,
    file_size: int,
) -> FetchRange:
    content = os.urandom(file_size)

    def fake_request(start: int, end: int) -> bytes:
        if start < 0 or end > file_size:
            raise ValueError('Request out of bounds')
        fetch_log.append((start, end))
        return content[start:end]

    return fake_request


@pytest.mark.parametrize('file_size', [2**20])
def test_file_read_cache(
    fetch_log: FetchLog,
    fetch_range: FetchRange,
    file_size: int,
) -> None:
    cache = FileReadCache(file_size, fetch_range)
    cache.read(100, 150)
    cache.read(5000, 6000)
    cache.read(8000, 9000)
    cache.read(20000, 20100)
    cache.read(30000, 32000)
    cache.read(10000, 100000)
    cache.read(file_size - 4, file_size)
    cache.read(file_size - 10000, file_size)
    cache.read(200000, 500000)
    cache.read(0, file_size)
    assert fetch_log == [
        (100, 8292),  # request coerced to min fetch size
        (8292, 16484),
        (20000, 28192),
        (30000, 38192),
        (16484, 20000),  # request spanning holes fills just holes
        (28192, 30000),
        (38192, 100000),
        (1040384, 1048576),  # request coerced to min fetch size backwards
        (1032192, 1040384),  # also expands backwards from a filled block
        (200000, 500000),
        (0, 100),
        (100000, 200000),
        (500000, 1032192),
    ]



import asyncio
import io
import logging

from collections.abc import Callable

from hctef.interval_tree import BinningIntervalTree, Interval

logger = logging.getLogger(__name__)


class AsyncFileReadCache:
    """
    An async-safe cache for file chunks that supports concurrent reads.

    Key Features:
    - Stores bytes | asyncio.Task[bytes] in intervals
    - Multiple reads of the same range share the same fetch task
    - Lock-free: uses asyncio's single-threaded execution model
    """

    def __init__(
        self,
        file_size: int,
        fetch_range: Callable[[int, int], asyncio.Task[bytes]],
        minimum_request_size: int = 8192,
    ) -> None:
        """
        Initialize async file read cache.

        Args:
            file_size: Total size of the file being cached
            fetch_range: Async function that returns a Task to fetch a byte range
            minimum_request_size: Minimum bytes to fetch in a single request
        """
        if file_size < 0:
            raise ValueError('File size cannot be less than zero')

        self.file_size = file_size
        self.cache = BinningIntervalTree[bytes | asyncio.Task[bytes]](
            bin_size=minimum_request_size,
        )
        self._fetcher = fetch_range
        self._minimum_request_size = minimum_request_size

    def clear(self):
        """
        Resets the cache, clearing all stored data and intervals.
        """
        logger.debug('Clearing all cached data')
        self.cache.clear()

    def _store_and_merge(self, start: int, end: int, chunk: bytes) -> None:
        """
        Stores a new chunk of data and merges it with any adjacent or
        overlapping chunks that contain completed bytes (not Tasks).
        """
        # Find adjacent/overlapping intervals that have completed (bytes only)
        overlapping = [
            iv
            for iv in self.cache.find_overlapping(start - 1, end + 1)
            if isinstance(iv.data, bytes)
        ]

        min_start = start
        max_end = end
        parts = {start: chunk}

        for iv in overlapping:
            min_start = min(min_start, iv.begin)
            max_end = max(max_end, iv.end)
            parts[iv.begin] = iv.data
            self.cache.remove(iv)

        merged_data = io.BytesIO()
        for offset in sorted(parts.keys()):
            merged_data.write(parts[offset])

        self.cache.add(Interval(min_start, max_end, merged_data.getvalue()))
        logger.debug(
            '--- CACHE UPDATE: Cache now contains: %s',
            [(iv.begin, iv.end, type(iv.data).__name__) for iv in sorted(self.cache)],
        )

    def _create_fetch_task(self, start: int, end: int) -> asyncio.Task[bytes] | None:
        """
        Creates a fetch task and stores it in the cache synchronously.

        Returns:
            Task to await, or None if range is already covered
        """
        # Apply minimum request size logic (same as sync version)
        more = self._minimum_request_size - (end - start)

        if more > 0:
            overlapping = self.cache.find_overlapping(
                end,
                min(self.file_size, end + self._minimum_request_size),
            )
            right_wall = overlapping[0].begin if overlapping else self.file_size
            end = min(end + more, right_wall)
            more = self._minimum_request_size - (end - start)

        if more > 0:
            overlapping = self.cache.find_overlapping(
                max(0, start - self._minimum_request_size),
                start,
                sort_by_end=True,
            )
            left_wall = overlapping[0].end if overlapping else 0
            start = max(start - more, left_wall)

        # Check if already covered (bytes or task)
        # Not needed in sync version due to no concurrency
        existing = self.cache.find_overlapping(start, end)
        if existing:
            logger.debug(
                'FETCH SKIPPED: bytes %s-%s (already in cache)',
                start,
                end - 1,
            )
            return None

        # Create task for the actual fetch operation
        async def do_fetch() -> bytes:
            chunk = await self._fetcher(start, end)
            # Remove task interval and store bytes
            self.cache.remove(task_interval)
            self._store_and_merge(start, end, chunk)
            return chunk

        task = asyncio.create_task(do_fetch())
        task_interval: Interval[bytes | asyncio.Task[bytes]] = Interval(
            start,
            end,
            task,
        )
        self.cache.add(task_interval)

        logger.debug(
            'FETCH INITIATED: bytes %s-%s (task created)',
            start,
            end - 1,
        )

        return task

    def _find_missing_ranges(self, start: int, end: int) -> list[Interval]:
        """
        Find 'holes' in the cache that need to be fetched.
        """
        missing: list[Interval] = []
        # All intervals (bytes or tasks) are considered "covered"
        relevant_intervals = sorted(self.cache.find_overlapping(start, end))
        pos = start

        for interval in relevant_intervals:
            if pos < interval.begin:
                missing.append(Interval(pos, interval.begin, None))
            pos = max(pos, interval.end)

        if pos < end:
            missing.append(Interval(pos, end, None))

        return missing

    async def read(self, start: int, end: int) -> bytes:
        """
        Reads a range of bytes asynchronously, utilizing the cache and
        fetching if necessary.
        """
        if end > self.file_size:
            raise ValueError('Read request extends beyond the end of the file.')

        logger.debug(
            'Read Request for bytes %s-%s',
            start,
            end - 1,
        )

        missing_intervals = self._find_missing_ranges(start, end)

        if not missing_intervals:
            logger.debug('CACHE HIT: All requested data already in cache.')
        else:
            logger.debug(
                'CACHE MISS: Missing intervals are: %s',
                [(iv.begin, iv.end) for iv in missing_intervals],
            )
            # Synchronously create all fetch tasks
            fetch_tasks = []
            for interval in missing_intervals:
                # Check if the interval (or part of it) has been filled
                # by a previous larger fetch in this same read() call.
                still_missing = self._find_missing_ranges(
                    interval.begin,
                    interval.end,
                )
                for gap in still_missing:
                    task = self._create_fetch_task(gap.begin, gap.end)
                    if task:
                        fetch_tasks.append(task)

        # Assemble result, awaiting any fetch tasks
        result_buffer = io.BytesIO()
        cached_chunks = sorted(self.cache.find_overlapping(start, end))

        for interval in cached_chunks:
            read_start = max(start, interval.begin)
            read_end = min(end, interval.end)

            # If this is still a task, await it to get bytes
            if isinstance(interval.data, asyncio.Task):
                data = await interval.data
            else:
                data = interval.data

            slice_start = read_start - interval.begin
            slice_end = read_end - interval.begin

            result_buffer.write(data[slice_start:slice_end])

        return result_buffer.getvalue()



import asyncio
import os

import pytest

from hctef.aio import AsyncHttpFile


@pytest.mark.asyncio
@pytest.mark.parametrize('parquet_file_name', ['alltypes_plain'])
async def test_async_http_file_basic(parquet_url: str) -> None:
    async with AsyncHttpFile(
        parquet_url,
        minimum_range_request_bytes=80,
        prefetch_bytes=100,
    ) as hf:
        # Test initial state
        assert hf.tell() == 0

        # Test reading a chunk
        data = await hf.read(100)
        assert len(data) == 100
        assert hf.tell() == 100

        # Test reading another chunk
        data2 = await hf.read(50)
        assert len(data2) == 50
        assert hf.tell() == 150

        # Test seeking from the beginning of the file (SEEK_SET)
        hf.seek(0, os.SEEK_SET)
        assert hf.tell() == 0
        data3 = await hf.read(10)
        assert len(data3) == 10
        assert hf.tell() == 10

        # Test seeking from the current position (SEEK_CUR)
        hf.seek(20, os.SEEK_CUR)
        assert hf.tell() == 30
        data4 = await hf.read(20)
        assert len(data4) == 20
        assert hf.tell() == 50

        # Test seeking from the end of the file (SEEK_END)
        hf.seek(-50, os.SEEK_END)
        file_size = hf.size
        assert hf.tell() == file_size - 50
        data5 = await hf.read(50)
        assert len(data5) == 50
        assert hf.tell() == file_size

        # Test reading past the end of the file
        assert await hf.read() == b''
        assert hf.tell() == file_size

        # Test seeking past the end of the file
        hf.seek(1000, os.SEEK_END)
        assert hf.tell() == file_size  # Position should be clamped
        assert await hf.read() == b''

        # Test seeking before the beginning of the file
        hf.seek(-file_size - 100, os.SEEK_END)
        assert hf.tell() == 0  # Position should be clamped


@pytest.mark.asyncio
@pytest.mark.parametrize('parquet_file_name', ['alltypes_plain'])
async def test_concurrent_reads_same_range(parquet_url: str) -> None:
    async with AsyncHttpFile(
        parquet_url,
        minimum_range_request_bytes=100,
        prefetch_bytes=0,  # Disable prefetch to control exactly what's fetched
    ) as hf:
        # Create cursors for concurrent reads
        c1 = hf.clone()
        c2 = hf.clone()

        # Issue 3 concurrent reads of the same range from position 0
        # All should share the same HTTP fetch
        results = await asyncio.gather(
            hf.read(100),  # Root cursor at position 0
            c1.read(100),  # Cursor 1 at position 0
            c2.read(100),  # Cursor 2 at position 0
        )

        # All should return the same data
        assert results[0] == results[1] == results[2]
        assert len(results[0]) == 100

        # Verify cache has the data
        cache = hf.cursor.ohf.cache
        cached_intervals = list(cache.cache.find_overlapping(0, 100))
        assert len(cached_intervals) > 0
        # Should be bytes, not a Task anymore
        assert isinstance(cached_intervals[0].data, bytes)


@pytest.mark.asyncio
@pytest.mark.parametrize('parquet_file_name', ['alltypes_plain'])
async def test_concurrent_reads_different_ranges(parquet_url: str) -> None:
    async with AsyncHttpFile(
        parquet_url,
        minimum_range_request_bytes=50,
        prefetch_bytes=0,
    ) as hf:
        # Create cursors for concurrent reads
        c1 = hf.clone()
        c2 = hf.clone()

        # Position each file handle differently
        hf.seek(0)
        c1.seek(200)
        c2.seek(400)

        # Issue concurrent reads from different positions
        results = await asyncio.gather(
            hf.read(50),  # Read bytes 0-50
            c1.read(50),  # Read bytes 200-250
            c2.read(50),  # Read bytes 400-450
        )

        # All should return different data (since they're from different positions)
        assert len(results[0]) == 50
        assert len(results[1]) == 50
        assert len(results[2]) == 50

        # The data should be different (different file positions)
        assert results[0] != results[1]
        assert results[1] != results[2]
        assert results[0] != results[2]


@pytest.mark.asyncio
@pytest.mark.parametrize('parquet_file_name', ['alltypes_plain'])
async def test_overlapping_concurrent_reads(parquet_url: str) -> None:
    async with AsyncHttpFile(
        parquet_url,
        minimum_range_request_bytes=50,
        prefetch_bytes=0,
    ) as hf:
        # First read to populate cache
        await hf.read(100)  # Bytes 0-100 now cached

        # Reset position
        hf.seek(0)

        # Issue overlapping reads: 0-150 and 50-200
        # First should use cache for 0-100, fetch 100-150
        # Second should use cache for 50-100, fetch 100-200
        c1 = hf.clone()
        c1.seek(50)

        results = await asyncio.gather(
            hf.read(150),  # Read bytes 0-150
            c1.read(150),  # Read bytes 50-200
        )

        assert len(results[0]) == 150
        assert len(results[1]) == 150

        # Verify the overlapping portion matches
        # results[0][50:150] should equal results[1][0:100]
        assert results[0][50:150] == results[1][0:100]


@pytest.mark.asyncio
@pytest.mark.parametrize('parquet_file_name', ['alltypes_plain'])
async def test_async_http_file_context_manager(parquet_url: str) -> None:
    hf = AsyncHttpFile(parquet_url, prefetch_bytes=0)

    # Should not be opened initially
    assert hf._cursor is None

    async with hf as f:
        # Should be opened
        assert f._cursor is not None
        await f.read(10)

    # Should be closed after context exits
    assert hf._cursor is None


@pytest.mark.asyncio
@pytest.mark.parametrize('parquet_file_name', ['alltypes_plain'])
async def test_read_without_open_raises(parquet_url: str) -> None:
    hf = AsyncHttpFile(parquet_url)

    with pytest.raises(ValueError, match='I/O operation on closed file'):
        await hf.read(10)


@pytest.mark.asyncio
@pytest.mark.parametrize('parquet_file_name', ['alltypes_plain'])
async def test_cache_deduplication_under_load(parquet_url: str) -> None:
    async with AsyncHttpFile(
        parquet_url,
        minimum_range_request_bytes=50,
        prefetch_bytes=0,
    ) as hf:
        # Create many cursors for concurrent reads of the same range
        # Only one fetch should occur
        num_reads = 20
        cursors = [hf.clone() for _ in range(num_reads - 1)]
        tasks = [hf.read(100)] + [c.read(100) for c in cursors]

        results = await asyncio.gather(*tasks)

        # All should return the same data
        assert all(r == results[0] for r in results)
        assert len(results[0]) == 100

        # Verify cache has the data and it's bytes (not a Task)
        cache = hf.cursor.ohf.cache
        cached_intervals = list(cache.cache.find_overlapping(0, 100))
        assert len(cached_intervals) > 0
        assert isinstance(cached_intervals[0].data, bytes)


@pytest.mark.asyncio
@pytest.mark.parametrize('parquet_file_name', ['alltypes_plain'])
async def test_clone_user_after_close(parquet_url: str) -> None:
    async with AsyncHttpFile(
        parquet_url,
        minimum_range_request_bytes=50,
        prefetch_bytes=0,
    ) as hf:
        cursor = hf.clone()
    with pytest.raises(RuntimeError, match='Session is closed'):
        await cursor.read(100)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass


@dataclass(eq=True, frozen=True)
class Interval[T]:
    """Represents an interval with a start, end, and associated data."""

    begin: int
    end: int
    data: T

    def __lt__(self, other) -> bool:
        if not isinstance(other, Interval):
            return NotImplemented
        return self.begin < other.begin


class BinningIntervalTree[T]:
    """
    An interval tree implementation that uses binning to speed up queries.
    This is a special-purpose implementation for FileReadCache.
    """

    def __init__(self, bin_size: int) -> None:
        if bin_size <= 0:
            raise ValueError('Bin size must be a positive integer')
        self.bin_size = bin_size
        self.bins: defaultdict[int, set[Interval[T]]] = defaultdict(set)
        self.all_intervals: set[Interval[T]] = set()

    def clear(self) -> None:
        """Removes all intervals from the tree."""
        self.bins.clear()
        self.all_intervals.clear()

    def add(self, interval: Interval[T]) -> None:
        """Adds a new interval to the tree."""
        self.all_intervals.add(interval)
        for bin_key in self._get_bin_keys_for_interval(interval):
            self.bins[bin_key].add(interval)

    def remove(self, interval: Interval[T]) -> None:
        """Removes a specific interval from the tree."""
        self.all_intervals.discard(interval)
        for bin_key in self._get_bin_keys_for_interval(interval):
            if bin_key in self.bins:
                self.bins[bin_key].discard(interval)

    def __iter__(self):
        return iter(self.all_intervals)

    def find_overlapping(
        self,
        start: int,
        end: int,
        sort_by_end: bool = False,
    ) -> list[Interval]:
        """Finds all intervals that overlap with the given range."""
        if start >= end:
            return []

        candidates: set[Interval[T]] = set()
        for bin_key in self._get_bin_keys_for_range(start, end):
            candidates.update(self.bins.get(bin_key, set()))

        overlapping = {iv for iv in candidates if iv.begin < end and iv.end > start}

        if sort_by_end:
            return sorted(overlapping, key=lambda iv: iv.end, reverse=True)
        return sorted(overlapping)

    def _get_bin_keys_for_interval(self, interval: Interval) -> range:
        start_bin = interval.begin // self.bin_size
        end_bin = (interval.end - 1) // self.bin_size
        return range(start_bin, end_bin + 1)

    def _get_bin_keys_for_range(self, start: int, end: int) -> range:
        start_bin = start // self.bin_size
        end_bin = (end - 1) // self.bin_size
        return range(start_bin, end_bin + 1)



from __future__ import annotations

import asyncio

from typing import Literal, Self

try:
    import aiohttp
except ImportError:
    raise ImportError(
        'Must install hctef with `[async]` extra to get necessary dependencies',
    ) from None

from hctef.exceptions import HctefNetworkError, HctefUrlError

from .async_file_read_cache import AsyncFileReadCache


def _check_url(url: str) -> None:
    """
    Validate that URL is a valid HTTP/HTTPS URL.

    Args:
        url: URL to validate

    Raises:
        HctefUrlError: If URL doesn't start with http: or https:
    """
    if not url.startswith(('http:', 'https:')):
        raise HctefUrlError("URL must start with 'http:' or 'https:'")


class _OpenedAsyncHttpFile:
    """
    Internal class managing shared state for AsyncHttpFile.
    """

    def __init__(
        self,
        http_file: AsyncHttpFile,
        session: aiohttp.ClientSession,
        size: int,
    ) -> None:
        """
        Initialize opened HTTP file with pre-fetched async values.

        Args:
            http_file: The parent AsyncHttpFile instance
            session: Pre-created aiohttp session
            size: File size obtained via async HTTP request
        """
        self.http_file = http_file
        self.session = session
        self.size = size
        self.cache = AsyncFileReadCache(
            self.size,
            self._fetch_range,
            minimum_request_size=min(
                self.http_file._minimum_range_request_bytes,
                self.size,
            ),
        )

    @classmethod
    async def create(cls, http_file: AsyncHttpFile) -> Self:
        """
        Async factory method to create _OpenedAsyncHttpFile.

        Args:
            http_file: The parent AsyncHttpFile instance

        Returns:
            Fully initialized _OpenedAsyncHttpFile instance
        """
        session = aiohttp.ClientSession()
        size = await cls._get_file_size(session, http_file.url)
        return cls(http_file, session, size)

    @staticmethod
    async def _get_file_size(
        session: aiohttp.ClientSession,
        url: str,
    ) -> int:
        """
        Get total file size using async HTTP range request.

        Args:
            session: aiohttp session to use for request
            url: URL to fetch size for

        Returns:
            File size in bytes

        Raises:
            HctefNetworkError: If size cannot be determined
        """
        try:
            headers = {'Range': 'bytes=0-'}
            async with session.get(url, headers=headers) as response:
                content_range = response.headers.get('Content-Range')
                if content_range:
                    return int(content_range.split('/')[-1])

                # If no Content-Range header, server doesn't support ranges
                raise HctefNetworkError(
                    f'Server does not support range requests for {url}',
                )
        except Exception as e:
            raise HctefNetworkError(
                f'Cannot determine file size for {url}',
            ) from e

    def _fetch_range(self, start: int, end: int) -> asyncio.Task[bytes]:
        """
        Create an async task to fetch byte range using HTTP request.

        Args:
            start: Start byte position (inclusive)
            end: End byte position (exclusive)

        Returns:
            Asyncio task that will fetch the requested bytes
        """
        return asyncio.create_task(self._do_fetch_range(start, end))

    async def _do_fetch_range(self, start: int, end: int) -> bytes:
        """
        Actually fetch byte range using async HTTP request.

        Args:
            start: Start byte position (inclusive)
            end: End byte position (exclusive)

        Returns:
            Bytes fetched from the range

        Raises:
            HctefNetworkError: If range request fails
        """
        if start >= end or start < 0 or end > self.size:
            raise HctefUrlError(
                f'Invalid byte range: {start}-{end} (file size: {self.size})',
            )

        try:
            headers = {'Range': f'bytes={start}-{end - 1}'}
            async with self.session.get(
                self.http_file.url,
                headers=headers,
            ) as response:
                return await response.read()
        except RuntimeError:
            raise
        except Exception as e:
            raise HctefNetworkError(
                f'Failed to fetch bytes {start}-{end} from {self.http_file.url}',
            ) from e

    async def read(self, position: int, size: int | None = None, /) -> bytes:
        """
        Read bytes from a specific position without managing cursor state.

        Args:
            position: Starting byte position to read from
            size: Number of bytes to read (None for all remaining)

        Returns:
            Bytes read from the file
        """
        if size is None:
            size = self.size - position

        if size < 0:
            raise ValueError(f'Cannot read negative number of bytes, got: {size}')

        if size == 0:
            return b''

        start = position
        end = min(start + size, self.size)

        return await self.cache.read(start, end)

    async def close(self) -> None:
        """
        Close the file and session.
        """
        await self.session.close()


class AsyncHttpFileCursor:
    """
    Lightweight cursor for reading from AsyncHttpFile with independent position.
    """

    def __init__(self, opened_file: _OpenedAsyncHttpFile) -> None:
        """
        Create a cursor for reading from an opened HTTP file.

        Args:
            opened_file: The shared opened file state
        """
        self.ohf = opened_file
        self.position = 0

    @property
    def size(self) -> int:
        return self.ohf.size

    async def read(self, size: int | None = None, /) -> bytes:
        """
        Read bytes from current position asynchronously.

        Args:
            size: Number of bytes to read (None for all remaining)

        Returns:
            Bytes read from the file
        """
        data = await self.ohf.read(self.position, size)
        self.position += len(data)
        return data

    def seek(self, offset: int, whence: int = 0, /) -> int:
        """
        Change stream position (synchronous - no I/O).

        Args:
            offset: Byte offset
            whence: How to interpret offset (0=absolute, 1=relative, 2=from end)

        Returns:
            New absolute position
        """
        if whence == 0:  # Absolute position
            new_pos = offset
        elif whence == 1:  # Relative to current position
            new_pos = self.position + offset
        elif whence == 2:  # Relative to end
            new_pos = self.size + offset
        else:
            raise ValueError(f'Invalid whence value: {whence}')

        if new_pos < 0:
            new_pos = 0
        elif new_pos > self.size:
            new_pos = self.size

        self.position = new_pos
        return self.position

    def tell(self) -> int:
        """
        Get current stream position (synchronous - no I/O).

        Returns:
            Current byte position in file
        """
        return self.position

    def clone(self) -> AsyncHttpFileCursor:
        """
        Create a new sibling cursor with independent position.

        Returns:
            New cursor sharing cache and session but with independent position

        Raises:
            ValueError: If file is not opened
        """
        return AsyncHttpFileCursor(self.ohf)

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def seekable(self) -> bool:
        return True


class AsyncHttpFile:
    """
    Async file-like wrapper for HTTP URLs with concurrent read support.
    """

    def __init__(
        self,
        url: str,
        minimum_range_request_bytes: int = 8192,
        prefetch_bytes: int = 2**20,
        prefetch_direction: Literal['START', 'END'] = 'END',
    ) -> None:
        """
        Initialize async HTTP file wrapper.

        Args:
            url: HTTP/HTTPS URL for a file

        Keyword Args:
            minimum_range_request_bytes:
                Least number of bytes to request,
                except when filling cache gaps
            prefetch_bytes:
                How many bytes to request when opening the file.
                Set to 0 or less to disable prefetch. Default 1 MiB.
            prefetch_direction:
                Whether to prefetch from file start or file end.
                Possible values `START` or `END`.

        Raises:
            HctefUrlError: If URL is invalid
        """
        _check_url(url)
        self.url = url
        self._prefetch_bytes = prefetch_bytes
        self._prefetch_direction = prefetch_direction
        self._minimum_range_request_bytes = minimum_range_request_bytes
        self._cursor: AsyncHttpFileCursor | None = None

    @property
    def cursor(self) -> AsyncHttpFileCursor:
        if not self._cursor:
            raise ValueError('I/O operation on closed file')
        return self._cursor

    @property
    def size(self) -> int:
        return self.cursor.size

    async def open(self) -> Self:
        """
        Open the file asynchronously.

        Returns:
            Self for use in context manager
        """
        self._cursor = AsyncHttpFileCursor(await _OpenedAsyncHttpFile.create(self))

        prefetch_bytes = min(self._prefetch_bytes, self.size)
        if prefetch_bytes > 0 and self._prefetch_direction == 'START':
            await self.read(prefetch_bytes)
        elif prefetch_bytes > 0 and self._prefetch_direction == 'END':
            self.cursor.seek(prefetch_bytes, 2)
            await self.read(prefetch_bytes)

        self.cursor.seek(0)

        return self

    def clone(self) -> AsyncHttpFileCursor:
        """
        Create a new cursor for concurrent reads.

        Returns:
            New cursor sharing cache but with independent position

        Raises:
            ValueError: If file is not opened
        """
        return self.cursor.clone()

    async def close(self) -> None:
        """
        Close the file and release resources.
        """
        if self._cursor:
            await self._cursor.ohf.close()
        self._cursor = None

    async def __aenter__(self) -> Self:
        """Async context manager entry."""
        return await self.open()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Async context manager exit."""
        await self.close()

    def __repr__(self) -> str:
        if self._cursor:
            return (
                f'AsyncHttpFile(url={self.url!r}, opened=True, '
                f'size={self.size}, pos={self._cursor.position})'
            )
        return f'AsyncHttpFile(url={self.url!r}, opened=False)'

    async def read(self, size: int | None = None, /) -> bytes:
        """
        Read bytes from current position asynchronously.

        Args:
            size: Number of bytes to read (None for all remaining)

        Returns:
            Bytes read from the file
        """
        return await self.cursor.read(size)

    def seek(self, offset: int, whence: int = 0, /) -> int:
        """
        Change stream position (synchronous - no I/O).

        Args:
            offset: Byte offset
            whence: How to interpret offset (0=absolute, 1=relative, 2=from end)

        Returns:
            New absolute position
        """
        return self.cursor.seek(offset, whence)

    def tell(self) -> int:
        """
        Get current stream position (synchronous - no I/O).

        Returns:
            Current byte position in file
        """
        return self.cursor.tell()

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def seekable(self) -> bool:
        return True



import os

import pytest

from hctef import HttpFile


@pytest.mark.parametrize('parquet_file_name', ['alltypes_plain'])
def test_http_file(parquet_url: str) -> None:
    with HttpFile(
        parquet_url,
        minimum_range_request_bytes=80,
        prefetch_bytes=100,
    ) as hf:
        # Test initial state
        assert hf.tell() == 0

        # Test reading a chunk
        data = hf.read(100)
        assert len(data) == 100
        assert hf.tell() == 100

        # Test reading another chunk
        data2 = hf.read(50)
        assert len(data2) == 50
        assert hf.tell() == 150

        # Test seeking from the beginning of the file (SEEK_SET)
        hf.seek(0, os.SEEK_SET)
        assert hf.tell() == 0
        data3 = hf.read(10)
        assert len(data3) == 10
        assert hf.tell() == 10

        # Test seeking from the current position (SEEK_CUR)
        hf.seek(20, os.SEEK_CUR)
        assert hf.tell() == 30
        data4 = hf.read(20)
        assert len(data4) == 20
        assert hf.tell() == 50

        # Test seeking from the end of the file (SEEK_END)
        hf.seek(-50, os.SEEK_END)
        file_size = hf._ohf._size
        assert hf.tell() == file_size - 50
        data5 = hf.read(50)
        assert len(data5) == 50
        assert hf.tell() == file_size

        # Test reading past the end of the file
        assert hf.read() == b''
        assert hf.tell() == file_size

        # Test seeking past the end of the file
        hf.seek(1000, os.SEEK_END)
        assert hf.tell() == file_size  # Position should be clamped
        assert hf.read() == b''

        # Test seeking before the beginning of the file
        hf.seek(-file_size - 100, os.SEEK_END)
        assert hf.tell() == 0  # Position should be clamped
